import asyncio
import re
from contextlib import AbstractContextManager
from gettext import gettext as _
from types import TracebackType
from typing import Callable, Literal, Optional, ParamSpec, Protocol, Sequence, Type, TypeVar

from babel.messages.frontend import CommandLineInterface
from textual.notifications import SeverityLevel
from textual.widget import Widget

from .log import Logger

R = TypeVar("R")
P = ParamSpec("P")


class NotifyProtocol(Protocol):
    def notify(
        self,
        message: str,
        *,
        title: str = "",
        severity: SeverityLevel = "information",
        timeout: float | None = None,
        markup: bool = True,
    ): ...


class NotifyException(AbstractContextManager["NotifyException"]):
    """A context manager that notifies the user of any exceptions that occur within its block."""

    def __init__(self, notifier: NotifyProtocol):
        """Initialize the NotifyException context manager.

        Args:
            notifier (NotifyProtocol): An object with a notify method to send notifications.
        """
        self.notifier = notifier
        self.logger = Logger()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_instance: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        if exc_instance:
            self.notifier.notify(
                _("An error occurred: {error}").format(error=str(exc_instance)),
                title=_("⛔ Unexpected Error"),
                severity="error",
                timeout=None,
                markup=False,
            )
            self.logger.error(
                "An error occurred",
                extra={
                    "context": "NotifyException.__exit__",
                    "error": str(exc_instance),
                    "traceback": str(traceback),
                    "type": str(exc_type),
                },
            )
            return True


def apply_styles(
    widget: Widget,
    vertical: Literal["middle", "top", "bottom"] = "middle",
    horizontal: Literal["center", "left", "right"] = "center",
    width: str | int | None = None,
    height: str | int | None = None,
) -> Widget:
    """Center the content of a Textual widget both horizontally and vertically.

    Args:
        widget (Widget): The widget whose content is to be centered.
    """
    widget.styles.content_align = (horizontal, vertical)
    widget.styles.align = (horizontal, vertical)
    if width is not None:
        widget.styles.width = width
    if height is not None:
        widget.styles.height = height
    return widget


def correct_translation(msgid: str, translation: str) -> str:
    """Correct common issues in the translated text.

    This includes fixing placeholders, extra spaces, and punctuation spacing.

    Args:
        msgid (str): The original message ID with placeholders.
        translation (str): The translated text to be corrected.
    Returns:
        str: The corrected translation.
    """
    placeholders = re.findall(r"\{[^}]+\}", msgid)
    for ph in placeholders:
        translation = re.sub(r"\{[^}]+\}", ph, translation, count=1)

    translation = re.sub(r"\s+", " ", translation)
    translation = re.sub(r"\s+([.,;:!?%\)])", r"\1", translation)
    translation = re.sub(r"\(\s+", "(", translation)
    translation = re.sub(r"\s+-\s*(\w)", r"-\1", translation)
    return translation.strip()


def escape_control_chars(text: str) -> str:
    """Escape control characters using character class pattern"""

    def replace_func(match: re.Match[str]) -> str:
        char = match.group(0)
        escape_map = {
            "\n": "\\n",  # new line
            "\r": "\\r",  # carriage return
            "\t": "\\t",  # tab
            "\b": "\\b",  # backspace
            "\f": "\\f",  # form feed
            "\v": "\\v",  # vertical tab
            "\a": "\\a",  # bell/alert
            "\\": "\\\\",  # backslash
            "\0": "\\0",  # null character
        }
        return escape_map.get(char, f"\\x{ord(char):02x}")

    return re.sub(r"[\n\r\t\b\f\v\a\\\x00]", replace_func, text)


def run_babel_cmd(args: Sequence[str]):
    """Run a Babel command with the given arguments."""
    cli = CommandLineInterface()
    print("Running Babel command with args:", args)
    cli.run(["pybabel", *args])  # pyright: ignore[reportUnknownMemberType]


def unescape_control_chars(text: str) -> str:
    """Unescape control chars including hex notation"""

    def replace(match: re.Match[str]) -> str:
        esc = match.group(0)
        if esc == r"\n":
            return "\n"
        if esc == r"\r":
            return "\r"
        if esc == r"\t":
            return "\t"
        if esc == r"\b":
            return "\b"
        if esc == r"\f":
            return "\f"
        if esc == r"\v":
            return "\v"
        if esc == r"\a":
            return "\a"
        if esc == r"\0":
            return "\0"
        if esc == r"\\":
            return "\\"
        if esc.startswith(r"\x"):
            return chr(int(esc[2:], 16))
        return esc  # fallback

    return re.sub(r"\\n|\\r|\\t|\\b|\\f|\\v|\\a|\\0|\\\\|\\x[0-9a-fA-F]{2}", replace, text)


async def wait_for_element(
    query_func: Callable[P, R],
    timeout: float = 5.0,
    interval: float = 0.1,
    *args: P.args,
    **kwargs: P.kwargs,
) -> R:
    """Wait for a UI element to be available by repeatedly calling the query function.

    Args:
        query_func (Callable[P, R]): The function to query the UI element.
        timeout (float): Maximum time to wait for the element.
        interval (float): Time between successive queries.
        *args: Positional arguments for the query function.
        **kwargs: Keyword arguments for the query function.

    Returns:
        R: The result from the query function.
    Raises:
        ValueError: If the element is not found within the timeout period.
    """
    total_time = 0.0
    while total_time < timeout:
        try:
            return query_func(*args, **kwargs)
        except Exception:
            await asyncio.sleep(interval)
            total_time += interval

    raise ValueError("UI element not found within timeout")



from gettext import gettext as _

from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container, Horizontal
from textual.screen import ModalScreen
from textual.widgets import Button, Footer, Header, Static

from ..log import Logger
from ..utils import apply_styles, wait_for_element


class ConfirmInevitable(ModalScreen[bool]):
    """A simple confirmation dialog modal."""

    BINDINGS = [
        Binding(key="left", action="focus_ok", description=_("Cancel"), show=False),
        Binding(key="right", action="focus_cancel", description=_("OK"), show=False),
    ]

    @property
    def logger(self) -> Logger:
        """Return the application logger."""
        return getattr(
            self.app,  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
            "logger",
        )

    def compose(self) -> ComposeResult:
        """Compose the modal's layout."""
        yield Header()
        yield apply_styles(
            Container(
                apply_styles(Static(_("Are you sure? This action cannot be undone.")), width="1fr"),
                apply_styles(
                    Horizontal(
                        Button(_("OK"), id="ok", variant="default"),
                        Button(_("Cancel"), id="cancel", variant="primary"),
                    ),
                    width="1fr",
                    height="1fr",
                ),
            ),
            width="1fr",
            height="1fr",
        )
        yield Footer()

    async def on_mount(self):
        """Focus the Cancel button when the modal is mounted."""
        (await wait_for_element(self.query_one, selector="#cancel", expect_type=Button)).focus()

    async def on_button_pressed(self, event: Button.Pressed):
        """Handle button press events."""
        if event.button.id == "ok":
            self.logger.debug(
                "User confirmed inevitable action.", extra={"context": "ConfirmInevitable.on_button_pressed"}
            )
            self.dismiss(True)
        elif event.button.id == "cancel":
            self.logger.debug(
                "User canceled inevitable action.", extra={"context": "ConfirmInevitable.on_button_pressed"}
            )
            self.dismiss(False)

    async def action_focus_ok(self):
        """Focus the OK button."""
        (await wait_for_element(self.query_one, selector="#ok", expect_type=Button)).focus()

    async def action_focus_cancel(self):
        """Focus the Cancel button."""
        (await wait_for_element(self.query_one, selector="#cancel", expect_type=Button)).focus()



from enum import Enum
from gettext import gettext as _
from pathlib import Path
from typing import Any, Dict, NamedTuple, Tuple

import polib


class POFileHandler:
    def __init__(self, po_path: Path):
        """Initialize the POFileHandler with the given PO file path.

        Args:
            po_path (Path): Path to the PO file.
        """
        self.pofile_path = po_path
        self.pofile = po_path  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]

    @property
    def pofile_path(self) -> Path:
        """Path to the PO file."""
        return self._po_path

    @pofile_path.setter
    def pofile_path(self, value: Path):
        """Set the PO file path."""
        self._po_path = value

    @property
    def pofile(self) -> polib.POFile:
        """Return the loaded PO file."""
        return self._po  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType, reportReturnType]

    @pofile.setter
    def pofile(self, value: polib.POFile | Path) -> None:
        """Set the PO file."""
        if isinstance(value, Path):
            file = polib.pofile(str(value))  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
            if isinstance(file, polib.POFile):
                self._po = file
            else:
                raise ValueError(f"Failed to load PO file from path: {value}")
        else:
            self._po = value


class SubCommand(NamedTuple):
    """A sub-command of the application."""

    name: str
    description: str
    requires_ui: bool
    default_check: bool

    @property
    def selection_list_item(self) -> Tuple[str, str, bool]:
        """Return a tuple suitable for use in a SelectionList item."""
        return (self.description, self.name, self.default_check)


class AppSubCommand(Enum):
    EXTRACT_UPDATE = SubCommand("extract_update", _("Extract and Update"), False, True)
    TRANSLATE = SubCommand("translate", _("Translate"), True, False)
    REVIEW = SubCommand("review", _("Review"), True, False)
    COMPILE = SubCommand("compile", _("Compile"), False, True)


class TableCell(NamedTuple):
    """A cell in the PO review table."""

    row_no: str
    type: str
    msgid: str
    msgstr: str
    msgstr: str
    msgstr: str


class SingletonType(type):
    """A metaclass for singleton classes."""

    _instances: Dict[str, object] = {}

    def __new__(cls, name: str, bases: Tuple[type, ...], attrs: Dict[str, Any]) -> Any:
        if name not in cls._instances:
            instance = super().__new__(cls, name, bases, attrs)
            cls._instances[name] = instance
        return cls._instances[name]



import argparse
import sys
import tomllib
from pathlib import Path
from typing import Any, Dict, List

import jmespath
from pydantic import BaseModel, Field

from .log import Logger

logger = Logger()


class CliArgument(dict[str, Any]):
    pass


class Config(BaseModel):
    """unbabelizer loads configuration from pyproject.toml and command line arguments."""

    author: str = Field(
        description="Author of the project",
        json_schema_extra={"pyproject.toml": "project.authors[0].name", "argparse": "--author"},
    )
    email: str = Field(
        description="Email of the author",
        json_schema_extra={"pyproject.toml": "project.authors[0].email", "argparse": "--email"},
    )
    version: str = Field(
        description="Version of the project",
        json_schema_extra={"pyproject.toml": "project.version", "argparse": "--version"},
    )
    title: str = Field(
        description="Title of the project", json_schema_extra={"pyproject.toml": "project.name", "argparse": "--title"}
    )
    locale_dir: Path = Field(
        default=Path("locale"),
        description="Directory for locale files",
        json_schema_extra={"pyproject.toml": "tool.unbabelizer.locale_dir", "argparse": "--locale-dir"},
    )
    input_paths: List[Path] = Field(
        default=[Path.cwd()],
        description="Paths to search for source files",
        json_schema_extra={"pyproject.toml": "tool.unbabelizer.input_paths", "argparse": "--input-paths"},
    )
    exclude_patterns: List[str] = Field(
        default=[],
        description="Paths to exclude from searching",
        json_schema_extra={"pyproject.toml": "tool.unbabelizer.exclude_patterns", "argparse": "--exclude-patterns"},
    )
    src_lang: str = Field(
        default="en",
        description="Source language code",
        json_schema_extra={"pyproject.toml": "tool.unbabelizer.src_lang", "argparse": "--src-lang"},
    )
    dest_lang: List[str] = Field(
        description="Destination language code",
        json_schema_extra={"pyproject.toml": "tool.unbabelizer.dest_lang", "argparse": "--dest-lang"},
    )
    domain: str = Field(
        default="messages",
        description="Domain for the .po files",
        json_schema_extra={"pyproject.toml": "tool.unbabelizer.domain", "argparse": "--domain"},
    )
    mapping_file: str = Field(
        default="[python: **.py]\nencoding = utf-8\n",
        description="Content of the Babel mapping file",
        json_schema_extra={"pyproject.toml": "tool.unbabelizer.mapping_file_content", "argparse": "--mapping-file"},
    )
    line_width: int = Field(
        default=120,
        description="Line width for .po files",
        json_schema_extra={
            "pyproject.toml": "tool.unbabelizer.line_width",
            "argparse": "--line-width",
        },
    )

    @classmethod
    def source_cli_args(cls, args: List[str]) -> Dict[str, Any]:
        """Parse command line arguments to create a Config instance."""
        parser = argparse.ArgumentParser(description="unbabelizer configuration")
        for name, field in cls.model_fields.items():
            schema = (  # pyright: ignore[reportUnknownVariableType]
                field.json_schema_extra  # pyright: ignore[reportUnknownMemberType]
                if isinstance(field.json_schema_extra, dict)
                else {}
            )
            flag = schema.get(  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
                "argparse", f"--{name.replace('_', '-')}"
            )
            parser.add_argument(
                flag,  # pyright: ignore[reportArgumentType]
                help=field.description or "",
                default=None,
                nargs=None if field.annotation is not list else "+",
                type=field.annotation if field.annotation is not None else str,
            )

        parsed_args = parser.parse_args(args)
        field_values = {k: v for k, v in vars(parsed_args).items() if v is not None}
        logger.debug("Parsed CLI args", extra={"context": "Config.source_cli_args", "data": field_values})
        return field_values

    @classmethod
    def source_pyproject_toml(cls, file_path: Path) -> Dict[str, Any]:
        """Load configuration from pyproject.toml using jmespath."""
        if not file_path.exists() or not file_path.is_file():
            return {}

        pyproject_data = tomllib.loads(file_path.read_text())
        logger.debug(
            "Loaded pyproject.toml data:", extra={"context": "Config.source_pyproject_toml", "data": pyproject_data}
        )
        field_values: Dict[str, Any] = {}
        for name, field in cls.model_fields.items():
            schema = (  # pyright: ignore[reportUnknownVariableType]
                field.json_schema_extra  # pyright: ignore[reportUnknownMemberType]
                if isinstance(field.json_schema_extra, dict)
                else {}
            )
            jmespath_query = schema.get(  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
                "pyproject.toml", name
            )
            value = jmespath.search(  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
                jmespath_query, pyproject_data
            )
            logger.debug(
                "JMESPath query result",
                extra={  # pyright: ignore[reportUnknownArgumentType]
                    "context": "Config.source_pyproject_toml",
                    "field": name,
                    "query": jmespath_query,
                    "value": value,
                },
            )
            field_values[name] = value

        field_values = {k: v for k, v in field_values.items() if v is not None}
        logger.debug(
            "Extracted config data from pyproject.toml",
            extra={"context": "Config.source_pyproject_toml", "data": field_values},
        )
        return field_values

    @classmethod
    def build(cls, args: List[str] | None) -> "Config":
        """Validate and create a Config instance from source data."""
        config_data = cls.source_pyproject_toml(Path("pyproject.toml"))
        config_data.update(cls.source_cli_args(args or sys.argv[1:]))
        logger.info("Validated config data", extra={"context": "Config.build", "data": config_data})
        return cls.model_validate(config_data)
